"""Main module."""
import os
import idzip
from tqdm import tqdm
from pathlib import Path
import threading
import concurrent.futures
import mmap
import struct

class PyLineSeek:
    """
    A database-like class for random-access to file lines.
    
    This class optionally compresses the file using idzip, writes an index of 
    byte offsets (stored as 8-byte unsigned integers), and supports concurrent
    and memory-mapped (mmap) read access.
    
    Parameters:
      filepath (str or Path): Path to the input file.
      compress (bool): If True, compress the file (resulting file gets '.dz' extension).
      header (bool): If True, the first line is treated as a header.
      load_index (bool): If True, load all byte offsets into memory.
      use_mmap (bool): If True and the file is uncompressed, use memory mapping for faster access.
    """
    ONE_MB = 1048576
    TEN_MB = 10 * ONE_MB
    HUNDRED_MB = 100 * ONE_MB

    def __init__(self, filepath, compress=True, header=False, load_index=False, use_mmap=False):
        self.filepath = Path(filepath)
        self.compress = compress
        self.header = header
        self.load_index = load_index
        self.use_mmap = use_mmap

        # Determine the output file.
        if self.compress:
            # Append .dz to the original suffix.
            self.output_filepath = self.filepath.with_suffix(self.filepath.suffix + ".dz")
        else:
            self.output_filepath = self.filepath

        self.output_directory = self.output_filepath.parent
        # Name the byte offsets file with a binary extension.
        self.byteoffset_filename = self.output_directory / "byteoffsets.bin"

        # Create a thread-local storage for file handles.
        self._thread_local = threading.local()

        # Get the number of lines from a pre-existing "numlines.txt" file in the output directory.
        numlines_file = self.output_directory / "numlines.txt"
        self.numlines = self._get_numlines(numlines_file)

        # If compression is enabled, ensure the compressed file exists.
        if self.compress and not self.output_filepath.exists():
            self.compress_file()

        # Open the main file (optionally with mmap).
        self._open_main_file()

        # If no index exists, write the byte offsets.
        if not self.byteoffset_filename.exists():
            self.write_byte_offsets()

        # If desired, load the index into memory.
        if self.load_index:
            self._load_offsets()

    def _get_numlines(self, numlines_file: Path) -> int:
        """
        Read and return the number of lines from a numlines.txt file.
        """
        with numlines_file.open("r") as f:
            return int(f.read().strip())

    def _open_main_file(self):
        """
        Open the main file for reading.
        
        If `use_mmap` is True and the file is uncompressed, memory-map the file.
        Otherwise, open it normally using the appropriate file opener.
        """
        if self.use_mmap and not self.compress:
            # Open the file normally so it can be memory-mapped.
            self._main_file_handle = open(self.output_filepath, "rb")
            self.main_file = mmap.mmap(self._main_file_handle.fileno(), 0, access=mmap.ACCESS_READ)
        else:
            # Use the idzip open for compressed files or standard open for uncompressed.
            self.main_file = (idzip.open(self.output_filepath, "rb")
                              if self.compress
                              else open(self.output_filepath, "rb"))
            self._main_file_handle = None

    def _get_thread_file(self):
        """
        Return a file-like object for the current thread.
        
        For mmap (read-only) files the same object is reused since it is thread-safe.
        Otherwise, a new file handle is opened per thread.
        """
        if hasattr(self._thread_local, "file"):
            return self._thread_local.file
        else:
            if self.use_mmap and not self.compress:
                # Sharing the mmap object is safe for concurrent read-only access.
                self._thread_local.file = self.main_file
            else:
                # Open a new file handle for this thread.
                self._thread_local.file = (idzip.open(self.output_filepath, "rb")
                                            if self.compress
                                            else open(self.output_filepath, "rb"))
            return self._thread_local.file

    def compress_file(self):
        """
        Compress the input file into a .dz file using idzip.
        """
        if not self.compress:
            print("Compression not enabled.")
            return

        # Compress in chunks.
        with self.filepath.open("rb") as infile, \
             idzip.IdzipFile(self.output_filepath, "wb", sync_size=10*self.HUNDRED_MB) as writer:
            chunk_size = self.HUNDRED_MB
            file_size = self.filepath.stat().st_size
            iterations = -(-file_size // chunk_size)  # Ceiling division.
            for _ in tqdm(range(iterations), desc="Compressing", unit="MB"):
                data = infile.read(chunk_size)
                if not data:
                    break
                writer.write(data)
        print(f"'{self.output_filepath}' has been created!")

    def write_byte_offsets(self):
        """
        Write the byte offsets for each line to a binary file.
        
        Each offset is stored as an 8-byte unsigned integer.
        The first offset is 0, and subsequent offsets correspond to the file pointer 
        after reading each line.
        """
        offsets = []
        if self.use_mmap and not self.compress:
            # If using mmap, we can work directly with the entire file content.
            data = self.main_file[:]
            offset = 0
            offsets.append(offset)
            # Using splitlines(keepends=True) to get line boundaries.
            for line in data.splitlines(keepends=True):
                offset += len(line)
                offsets.append(offset)
        else:
            # Read line by line from the current file handle.
            f = self.main_file
            f.seek(0)
            offsets.append(0)
            while True:
                pos = f.tell()
                line = f.readline()
                if not line:
                    break
                offsets.append(f.tell())

        # Write all offsets in binary form.
        with self.byteoffset_filename.open("wb") as index_file:
            for off in offsets:
                index_file.write(struct.pack("<Q", off))

        # Reset the main file pointer.
        if not (self.use_mmap and not self.compress):
            self.main_file.seek(0)
        # If the index is loaded in memory, update it.
        if self.load_index:
            self.offsets = offsets

    def _load_offsets(self):
        """
        Load the byte offsets from the binary index file into memory.
        """
        self.offsets = []
        with self.byteoffset_filename.open("rb") as index_file:
            offset_data = index_file.read()
            count = len(offset_data) // 8  # 8 bytes per offset.
            for i in range(count):
                (off,) = struct.unpack("<Q", offset_data[i*8:(i+1)*8])
                self.offsets.append(off)

    def __len__(self):
        """
        Return the number of lines in the file.
        """
        return self.numlines

    def __getitem__(self, key, workers=1):
        """
        Retrieve a single line or a slice of lines.
        
        For slices, concurrent access is used (if workers > 1). Note that you may
        pass a custom 'workers' argument if desired.
        
        Parameters:
          key (int or slice): The line index or slice.
          workers (int): Number of threads to use for concurrent reads (default: 1).
        Returns:
          A string (for int index) or a list of strings (for slice).
        """
        if workers == -1:
            workers = os.cpu_count()

        if isinstance(key, slice):
            start = key.start or 0
            stop = key.stop or self.numlines
            step = key.step or 1
            indices = list(range(start, stop, step))
            if workers > 1:
                with concurrent.futures.ThreadPoolExecutor(max_workers=workers) as executor:
                    lines = list(executor.map(self.get_line, indices))
                return lines
            else:
                return [self.get_line(i) for i in indices]
        elif isinstance(key, int):
            if key < 0:  # Handle negative indexing.
                key += self.numlines
            return self.get_line(key)
        else:
            raise TypeError("Index must be int or slice.")

    def get_line(self, line_number: int) -> str:
        """
        Retrieve a specific line from the file.
        
        Parameters:
          line_number (int): Zero-based index of the desired line.
        Returns:
          str: The decoded content of the line.
        """
        byte_offset = self.get_byte_offset(line_number)
        f = self._get_thread_file()
        f.seek(byte_offset, os.SEEK_SET)
        return f.readline().decode()

    def get_byte_offset(self, line_number: int) -> int:
        """
        Retrieve the byte offset for a given line.
        
        Adjusts the line number if a header is present.
        
        Parameters:
          line_number (int): Zero-based index of the line.
        Returns:
          int: The byte offset from the beginning of the file.
        """
        if self.header:
            line_number += 1
        if line_number >= len(self):
            raise IndexError("Line number out of range")
        if self.offsets is not None:
            # Use the in-memory cache.
            return self.offsets[line_number]
        else:
            # Read the 8-byte offset directly from the index file.
            with self.byteoffset_filename.open("rb") as index_file:
                index_file.seek(line_number * 8)
                data = index_file.read(8)
                if not data or len(data) < 8:
                    raise IndexError("Offset not found")
                (off,) = struct.unpack("<Q", data)
                return off

    def test_random_line(self, line_number: int):
        """
        Validate that the random-access retrieval of a line matches sequential iteration.
        
        Parameters:
          line_number (int): The line number to test.
        Raises:
          AssertionError: If the lines do not match.
        """
        line_from_get_line = self.get_line(line_number).strip()
        f = self._get_thread_file()
        f.seek(0)
        for i, line in enumerate(f):
            if i == line_number:
                line_from_iteration = line.decode().strip()
                break
        assert line_from_get_line == line_from_iteration, (
            f"Lines don't match! {line_from_get_line} != {line_from_iteration}"
        )
        print("Test passed!")

    def clear(self):
        """
        Remove generated compressed and index files.
        """
        if self.compress and self.output_filepath.exists():
            self.output_filepath.unlink()
            print(f"Removed {self.output_filepath}")
        if self.byteoffset_filename.exists():
            self.byteoffset_filename.unlink()
            print(f"Removed {self.byteoffset_filename}")

    def close(self):
        """
        Close all open file handles.
        """
        # Close the main file or mmap.
        if self.use_mmap and not self.compress:
            if self.main_file:
                self.main_file.close()
            if self._main_file_handle:
                self._main_file_handle.close()
        else:
            if self.main_file:
                self.main_file.close()
        # Close any thread-local file handles.
        if hasattr(self._thread_local, "file"):
            try:
                self._thread_local.file.close()
            except Exception:
                pass

    def __enter__(self):
        """
        Enter the runtime context related to this object.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit the runtime context and close open resources.
        """
        self.close()

    def __del__(self):
        """
        Destructor ensuring resources are freed.
        """
        self.close()
